/*
 * Interrupt and TLB miss handling support.
 *
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mips.h"
#include "paging.h"

.text
.extern interrupt_handler
.extern current_registers
.extern current_stack_pointer
.globl _tlb_entry
.globl _exc_entry
.globl _irq_entry
.globl _end_entries
.set noreorder

_tlb_entry:
	/* Get the bad address. */

	mfc0 $k0, CP0_ENTRYHI
	nop
	andi $k1, $k0, 0xff			/* ASID */

	/* For ASID == 0... */

	beqz $k1, _tlb_entry_direct
	nop

	/* Otherwise, load the page table entries. */

	andi $k1, $k0, 0xff			/* ASID */
	li $k0, page_table_task_size
	mul $k0, $k0, $k1			/* [ASID] (ASID * page_table_task_size) */
	li $k1, page_table_start		/* page_table */
	addu $k1, $k0, $k1			/* page_table[ASID] */

	mfc0 $k0, CP0_CONTEXT
	nop
	srl $k0, $k0, 1				/* use 8 byte - not 16 byte - entries */
	addu $k0, $k0, $k1			/* page_table[ASID][entry] */

	lw $k1, 0($k0)				/* page_table[ASID][entry][0] */
	mtc0 $k1, CP0_ENTRYLO0

	lw $k1, 4($k0)				/* page_table[ASID][entry][1] */
	mtc0 $k1, CP0_ENTRYLO1
						/* page size is 4KB */
	mtc0 $zero, CP0_PAGEMASK
	nop

	tlbwr
	nop

	eret
	nop

_tlb_entry_direct:
	/* Otherwise, just translate the address directly. */

	li $k1, 0xffffe000
	and $k0, $k0, $k1			/* VPN2 (8KB resolution) */
	srl $k0, $k0, 6				/* PFN (maintain 8KB resolution, bit 6 remaining zero) */
	ori $k0, $k0, 0x1e			/* flags */

	mtc0 $k0, CP0_ENTRYLO0
	ori $k0, $k0, 0x40			/* page size is 4KB (bit 6 set) */
	mtc0 $k0, CP0_ENTRYLO1
	nop					/* page size is 4KB */
	mtc0 $zero, CP0_PAGEMASK
	nop

	tlbwr
	nop

	eret
	nop

_exc_entry:
	/* Handle TLB refill exceptions. */

	mfc0 $k0, CP0_CAUSE
	li $k1, 0x0000007c
	and $k0, $k0, $k1			/* ExcCode << 2 */
	srl $k0, $k0, 2				/* ExcCode */
	addi $k1, $k0, -2			/* ExcCode == 2 */
	beqz $k1, _tlb_entry
	addi $k1, $k0, -3			/* ExcCode == 3 */
	beqz $k1, _tlb_entry
	nop
_fail:
	b _fail
	nop

_irq_entry:
	/* Set the ASID. */

	mtc0 $zero, CP0_ENTRYHI
	nop

	/* Obtain the kernel global offset table. */

	move $k0, $gp
	lui $gp, %hi(_GLOBAL_OFFSET_TABLE_)
	ori $gp, $gp, %lo(_GLOBAL_OFFSET_TABLE_)

	/* Obtain a store of registers for the current task. */

	la $k1, current_registers
	lw $k1, 0($k1)

	/* Save registers that the assembler wants to trash. */

	sw $t9, 100($k1)
	sw $k0, 104($k1)			/* old gp */
	sw $ra, 112($k1)

	/* Record the stack pointer. */

	la $k1, current_stack_pointer
	lw $k1, 0($k1)				/* &stack_pointers[current_task] */
	sw $sp, 0($k1)

	/* Switch to the kernel stack. */

	li $sp, 0x80070000

	/* Invoke the rest of the interrupt handling process. */

	la $k0, interrupt_handler
	jr $k0
	nop

_end_entries:

.set reorder
